"""
Custom exceptions for the License Manager Agent.
"""

from buzz import Buzz


class LicenseManagerAuthTokenError(Buzz):
    """Exception for backend connection issues."""


class LicenseManagerBackendConnectionError(Buzz):
    """Exception for backend connection issues."""


class LicenseManagerParseError(Buzz):
    """Exception for error during parsing of backend data."""


class LicenseManagerBackendVersionError(Buzz):
    """Exception for backend/agent version mismatches."""


class LicenseManagerEmptyReportError(Buzz):
    """Exception for empty report when no licenses added in backend."""


class LicenseManagerNonSupportedServerTypeError(Buzz):
    """Exception for entry with non supported server type."""


class LicenseManagerBadServerOutput(Buzz):
    """Exception for license server bad output."""


class CommandFailedToExecute(Buzz):
    """Exception for failed execution of command."""


class LicenseManagerFeatureConfigurationIncorrect(Buzz):
    """Exception for feature configuration incorrectly formatted."""


class LicenseManagerReservationFailure(Buzz):
    """Exception for failure during reservation management."""


class SqueueParserUnexpectedInputError(Buzz):
    """Unexpected squeue output."""


class ScontrolRetrievalFailure(Buzz):
    """Could not get SLURM data for job id."""



"""Provide utilities to be used by interfaces."""
import asyncio
import shlex
from typing import List, Optional

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute
from lm_agent.logs import logger


async def run_command(command_line_parts: List[str], stdin_str: Optional[str] = "") -> str:
    """
    Run a command using a subprocess shell.

    The command is passed as a list of strings, where each string is a part of the command.
    You can also provide a string to be passed as stdin to the process.

    Returns the output as string if the command succeeds.
    Raises CommandFailedToExecute exception if return code is not zero.
    """

    command_line = shlex.join(command_line_parts)
    stdin_bytes = stdin_str.encode(settings.ENCODING) if stdin_str else None

    proc = await asyncio.create_subprocess_shell(
        command_line,
        stdin=asyncio.subprocess.PIPE if stdin_bytes else None,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    try:
        stdout, _ = await asyncio.wait_for(proc.communicate(stdin_bytes), settings.TOOL_TIMEOUT)
        output = str(stdout, encoding=settings.ENCODING, errors="replace")

        if proc.returncode != 0:
            error_message = shlex.join(
                [
                    f"Command {command_line} failed!",
                    f"Error: {output}",
                    f"Return code: {proc.returncode}",
                ]
            )
            logger.error(error_message)
            raise CommandFailedToExecute(
                f"The command failed to execute, with return code {proc.returncode}."
            )
    except asyncio.TimeoutError:
        logger.error(f"Command {command_line} timed out after {settings.TOOL_TIMEOUT} seconds.")
        raise CommandFailedToExecute(
            f"The command failed to execute, timed out after {settings.TOOL_TIMEOUT} seconds."
        )

    return output



"""
Parser for FlexLM
"""
import re
from typing import Dict, Optional
from lm_agent.models import LicenseUsesItem


HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD})*"
NONWS = r"\S+"
NONCOLON = r"[^:]+"
INT = r"\d+"
NONPAREN = r"[^() ]+"
MMDD = rf"{INT}/{INT}"
HHMM = rf"{INT}:{INT}"
FEATURE_NAME = r"[A-Za-z0-9\-_:]+"


FEATURE_LINE = (
    r"^Users of "
    rf"(?P<feature>{FEATURE_NAME}):  "
    r"\(Total of "
    rf"(?P<total>{INT}) "
    r"licenses issued;  Total of "
    rf"(?P<used>{INT}) "
    rf"license(?P<plural>s)? in use\)$"
)

USAGE_LINE_1 = (
    r"^\s+"
    rf"(?P<user>{NONWS}) "
    rf"(?P<user_host>{HOSTNAME}) "
    rf"(?P<tty>{NONWS}) "
    rf"\((?P<version>{NONPAREN})\) "
    rf"\((?P<license_host>{HOSTNAME})/"
    rf"(?P<license_port>{INT}) "
    rf"(?P<license_pid>{INT})\), "
    r"start "
    rf"(?P<weekday>{NONWS}) "
    rf"(?P<date>{MMDD}) "
    rf"(?P<time>{HHMM}), "
    rf"(?P<tokens>{INT}) "
    r"licenses$"
)

USAGE_LINE_2 = (
    r"^\s+"
    rf"(?P<user>{NONWS}) "
    rf"(?P<user_host>{HOSTNAME}) "
    rf"(?P<tty>{NONWS}) "
    r"feature="
    rf"(?P<feature>{FEATURE_NAME}) "
    rf"\(v(?P<version>{NONPAREN})\) "
    rf"\((?P<license_host>{HOSTNAME})/"
    rf"(?P<license_port>{INT}) "
    rf"(?P<license_pid>{INT})\), "
    r"start "
    rf"(?P<weekday>{NONWS}) "
    rf"(?P<date>{MMDD}) "
    rf"(?P<time>{HHMM}), "
    rf"(?P<tokens>{INT}) "
    r"licenses$"
)


USAGE_LINE_3 = (
    r"^\s+"
    rf"(?P<user>{NONWS}) "
    rf"(?P<user_host>{HOSTNAME}) "
    rf"(?P<tty>{NONWS}) "
    rf"\(v(?P<version>{NONPAREN})\) "
    rf"\((?P<license_host>{HOSTNAME})/"
    rf"(?P<license_port>{INT}) "
    rf"(?P<license_pid>{INT})\), "
    r"start "
    rf"(?P<weekday>{NONWS}) "
    rf"(?P<date>{MMDD}) "
    rf"(?P<time>{HHMM})$"
)

USAGE_LINE_4 = (
    r"^\s+"
    rf"(?P<user>{NONWS}) "
    rf"(?P<user_host>{HOSTNAME}) "
    rf"(?P<tty>{HOSTNAME}) "
    rf"(?P<feature>{FEATURE_NAME}) "
    rf"\(v(?P<version>{NONPAREN})\) "
    rf"\((?P<license_host>{HOSTNAME})/"
    rf"(?P<license_port>{INT}) "
    rf"(?P<license_pid>{INT})\), "
    r"start "
    rf"(?P<weekday>{NONWS}) "
    rf"(?P<date>{MMDD}) "
    rf"(?P<time>{HHMM}), "
    rf"(?P<tokens>{INT}) "
    r"licenses$"
)

USAGE_LINE_5 = (
    r"^\s+"
    rf"(?P<user>{NONWS}) "
    rf"(?P<user_host>{HOSTNAME}) "
    rf"(?P<tty>{NONWS}) "
    rf"(?P<feature>{FEATURE_NAME}) "
    rf"\(v(?P<version>{NONPAREN})\) "
    rf"\((?P<license_host>{HOSTNAME})/"
    rf"(?P<license_port>{INT}) "
    rf"(?P<license_pid>{INT})\), "
    r"start "
    rf"(?P<weekday>{NONWS}) "
    rf"(?P<date>{MMDD}) "
    rf"(?P<time>{HHMM})$"
)

RX_FEATURE = re.compile(FEATURE_LINE)
RX_USAGE_LINE_1 = re.compile(USAGE_LINE_1)
RX_USAGE_LINE_2 = re.compile(USAGE_LINE_2)
RX_USAGE_LINE_3 = re.compile(USAGE_LINE_3)
RX_USAGE_LINE_4 = re.compile(USAGE_LINE_4)
RX_USAGE_LINE_5 = re.compile(USAGE_LINE_5)
USAGE_LINES = [RX_USAGE_LINE_1, RX_USAGE_LINE_2, RX_USAGE_LINE_3, RX_USAGE_LINE_4, RX_USAGE_LINE_5]


def parse_feature_line(line: str) -> Optional[Dict]:
    """
    Parse the feature line in the FlexLM output.
    Data we need:
    - ``feature``: license name
    - ``total``: total amount of licenses
    - ``used``: quantity of licenses being used
    """
    parsed_feature = RX_FEATURE.match(line)
    if parsed_feature is None:
        return None
    feature_data = parsed_feature.groupdict()

    return {
        "feature": feature_data["feature"].lower(),
        "total": int(feature_data["total"]),
        "used": int(feature_data["used"]),
    }


def parse_usage_line(line: str) -> Optional[LicenseUsesItem]:
    """
    Parse the usage line in the FlexLM output.
    Data we need:
    - ``username``: user who booked the license
    - ``lead_host``: host using the license
    - ``booked``: quantity of licenses being used

    There can be multiple formats for the data line, so we need to check which one matches.
    """

    for RX in USAGE_LINES:
        if parsed_data := RX.match(line):
            break

    if parsed_data is None:
        return None

    data = parsed_data.groupdict()

    return LicenseUsesItem(
        username=data["user"].lower(),
        lead_host=data["user_host"],
        booked=int(data.get("tokens", 1)),
    )


def parse(server_output: str) -> Dict:
    """
    Parse the FlexLM Output, using regext to match the lines we need:
    - ``feature line``: info about the license
    - ``data line``: info about the users using the license
    """
    parsed_data: dict = {}
    uses = []

    for line in server_output.splitlines():
        parsed_feature = parse_feature_line(line)

        if parsed_feature:
            parsed_data = parsed_feature
            continue

        parsed_data_line = parse_usage_line(line)

        if parsed_data_line:
            uses.append(parsed_data_line)

    if parsed_data:
        parsed_data["uses"] = uses

    return parsed_data



"""
Reconciliation functionality live here.
"""
from lm_agent.backend_utils.utils import (
    get_all_features_bookings_sum,
    get_cluster_configs_from_backend,
    get_cluster_jobs_from_backend,
)
from lm_agent.logs import logger
from lm_agent.services.license_report import update_features
from lm_agent.services.clean_jobs_and_bookings import clean_jobs_and_bookings
from lm_agent.workload_managers.slurm.cmd_utils import (
    get_all_features_cluster_values,
    return_formatted_squeue_out,
    squeue_parser,
)
from lm_agent.workload_managers.slurm.reservations import (
    create_or_update_reservation,
    scontrol_delete_reservation,
    scontrol_show_reservation,
)


async def reconcile():
    """Generate the report and reconcile the license feature token usage."""
    logger.debug("Starting reconciliation")

    # Generate report and update the backend
    logger.debug("Reconciling licenses in the backend")
    license_usage_info = await update_features()
    logger.debug("Backend licenses reconciliated")

    # Get cluster data
    configurations = await get_cluster_configs_from_backend()

    # Get cluster jobs
    jobs = await get_cluster_jobs_from_backend()

    # Get feature bookings sum
    all_features_bookings_sum = await get_all_features_bookings_sum()

    # Get license usage from the cluster
    all_features_cluster_value = await get_all_features_cluster_values()

    # Get squeue result from cluster
    squeue_output = squeue_parser(await return_formatted_squeue_out())

    # Clean jobs and bookings
    await clean_jobs_and_bookings(configurations, jobs, squeue_output, license_usage_info)

    reservation_data = []

    # Calculate how many licenses should be reserved for each license
    for license_data in license_usage_info:
        # Get license usage from license report
        product_feature = license_data.product_feature
        report_used = license_data.used
        report_total = license_data.total

        # Get booking information from backend
        booking_sum = all_features_bookings_sum[product_feature]

        # Get license server type and reserved from the configuration in the backend
        for configuration in configurations:
            for feature in configuration.features:
                if f"{feature.product.name}.{feature.name}" == product_feature:
                    license_server_type = configuration.type.value

        # Get license usage from the cluster
        slurm_used = all_features_cluster_value[product_feature]["used"]
        slurm_total = all_features_cluster_value[product_feature]["total"]

        """
        The reserved amount represents how many licenses are already in use:
        Either in the license server or booked for a job (bookings from other cluster as well).

        If the license has a reserved value (licenses exclusive for usage in desktop environments),
        the value will be decreased from the Slurm counter and will be checked at booking creation
        time. This implies that the value won't need to be added to the reservation.

        The reservation is not meant to be used by any user, it's a way to block usage of licenses.

        If the report total is 0, it means that the license is not available in the license server
        and it should be fully reserved to prevent jobs from running and crashing.
        """

        if report_total == 0:
            reservation_amount = slurm_total
        else:
            reservation_amount = report_used - slurm_used + booking_sum

        if reservation_amount < 0:
            reservation_amount = 0

        if reservation_amount > slurm_total:
            reservation_amount = slurm_total

        if reservation_amount:
            reservation_data.append(f"{product_feature}@{license_server_type}:{reservation_amount}")

    if reservation_data:
        logger.debug(f"Reservation data: {reservation_data}")

        # Create the reservation or update the existing one
        await create_or_update_reservation(",".join(reservation_data))
    else:
        logger.debug("No reservation needed")

        existing_reservation = await scontrol_show_reservation()
        if existing_reservation:
            logger.debug("Deleting existing reservation")
            await scontrol_delete_reservation()

    logger.debug("Reconciliation done")



"""
Common utilities for slurm commands.
"""
import os

from lm_agent.logs import logger
from lm_agent.workload_managers.slurm.cmd_utils import get_lead_host


async def get_job_context():
    """
    Get and return variables from the job environment.
    """
    ctxt = dict()
    try:
        ctxt = {
            "cluster_name": os.environ["SLURM_CLUSTER_NAME"],
            "job_id": os.environ["SLURM_JOB_ID"],
            "lead_host": await get_lead_host(os.environ["SLURM_JOB_NODELIST"]),
            "user_name": os.environ["SLURM_JOB_USER"],
            "job_licenses": os.environ.get("SLURM_JOB_LICENSES", ""),
        }

    except KeyError as e:
        # If not all keys could be assigned, then return non 0 exit status
        logger.error(
            f"All required environment variables were not set, missing: {e}. "
            "Expecting: SLURM_CLUSTER_NAME, SLURM_JOB_ID, SLURM_JOB_NODELIST, "
            "SLURM_JOB_USER, SLURM_JOB_LICENSES"
        )

    return ctxt



"""
Parser for DSLS.
"""
import csv
from typing import Optional

from lm_agent.models import LicenseUsesItem, ParsedFeatureItem


def parse_feature_dict(feature_dict: dict[str, str]) -> Optional[ParsedFeatureItem]:
    """
    Parse the feature dcit in the DSLS output.

    Data we need:
    - ``feature``: license name
    - ``total``: total amount of licenses
    - ``used``: quantity of licenses being used
    """
    if not all([feature_dict.get("Feature"), feature_dict.get("Count"), feature_dict.get("Inuse")]):
        return None

    return ParsedFeatureItem(
        feature=feature_dict["Feature"].lower(),
        total=int(feature_dict["Count"]),
        used=int(feature_dict["Inuse"]),
        uses=[],
    )


def parse_usage_dict(usage_dict: dict[str, str]) -> Optional[LicenseUsesItem]:
    """
    Parse the usage dict in the DSLS output.

    Data we need:
    - ``username``: user who booked the license
    - ``lead_host``: host using the license
    - ``booked``: quantity of licenses being used
    """
    if not all([usage_dict.get("User"), usage_dict.get("Host"), usage_dict.get("Tokens")]):
        return None

    return LicenseUsesItem(
        username=usage_dict["User"],
        lead_host=usage_dict["Host"].split()[0],
        booked=int(usage_dict["Tokens"]),
    )


def parse(server_output: str) -> dict[str, ParsedFeatureItem]:
    """
    Parse the output from the DSLS server.
    Data we need:
    - ``feature``: license name
    - ``count``: total amount of licenses
    - ``in_use``: quantity of licenses being use
    - ``uses``: list of users using the license

    Obs: if the feature is used by multiple users, each usage will result in a line
    with the same feature information, just changing the usage data.
    """
    parsed_data: dict = {}

    # Ignore the lines that contain information about the DSLS license server
    if "Warning" in server_output:  # Remove warning line with the information lines
        offset = 7
    else:
        offset = 6

    csv_data = csv.DictReader(server_output.splitlines()[offset:])

    for row in csv_data:
        parsed_feature = parse_feature_dict(row)
        if not parsed_feature:
            continue

        parsed_usage = parse_usage_dict(row)
        if parsed_usage:
            parsed_feature.uses.append(parsed_usage)

        if parsed_feature.feature in parsed_data:
            parsed_data[parsed_feature.feature].uses.extend(parsed_feature.uses)
        else:
            parsed_data[parsed_feature.feature] = parsed_feature

    return parsed_data



from typing import List

from pydantic import BaseModel, Field, PositiveInt

from lm_agent.constants import LicenseServerType, PRODUCT_FEATURE_RX


class LicenseUsesItem(BaseModel):
    """
    A list of usage information for a license.
    """

    username: str
    lead_host: str
    booked: int


class LicenseReportItem(BaseModel):
    """
    An item in a LicenseReport, a count of tokens for one product/feature.
    """

    feature_id: int
    product_feature: str = Field(..., pattern=PRODUCT_FEATURE_RX)
    used: int
    total: int
    uses: List[LicenseUsesItem] = []


class ParsedFeatureItem(BaseModel):
    """
    A report of the parsed license server output.
    """

    feature: str
    total: int
    used: int
    uses: List[LicenseUsesItem] = []


class LicenseBooking(BaseModel):
    """
    Structure to represent a license booking.
    """

    product_feature: str = Field(..., pattern=PRODUCT_FEATURE_RX)
    quantity: int


class LicenseBookingRequest(BaseModel):
    """
    Structure to represent a list of license bookings.
    """

    slurm_job_id: str
    username: str
    lead_host: str
    bookings: List[LicenseBooking] = []


class ExtractedBookingSchema(BaseModel):
    """
    Represents the booking from a job with the job information extracted.
    """

    booking_id: int
    job_id: int
    slurm_job_id: str
    username: str
    lead_host: str
    feature_id: int
    quantity: int


class ExtractedUsageSchema(BaseModel):
    """
    Representes the usage lines extracted from a feature report.
    """

    feature_id: int
    username: str
    lead_host: str
    quantity: int


class LicenseServerSchema(BaseModel):
    """
    License server response from the database.
    """

    id: int
    config_id: int
    host: str
    port: PositiveInt


class ProductSchema(BaseModel):
    """
    Represents a feature's product.
    """

    id: int
    name: str


class FeatureSchema(BaseModel):
    """
    Represents the features in a feature configuration.
    """

    id: int
    name: str
    product: ProductSchema
    config_id: int
    reserved: int
    total: int
    used: int
    booked_total: int


class ConfigurationSchema(BaseModel):
    """
    Represents the configuration for a set of features.
    """

    id: int
    name: str
    cluster_client_id: str
    features: List[FeatureSchema] = []
    license_servers: List[LicenseServerSchema] = []
    grace_time: int
    type: LicenseServerType


class BookingSchema(BaseModel):
    """
    Represents the booking of a feature.
    """

    id: int
    job_id: int
    feature_id: int
    quantity: int


class JobSchema(BaseModel):
    """
    Represents the jobs submitted in a cluster.
    """

    id: int
    slurm_job_id: str
    cluster_client_id: str
    username: str
    lead_host: str

    bookings: List[BookingSchema] = []



"""LS-Dyna license server interface."""
import typing

from lm_agent.models import LicenseServerSchema, LicenseReportItem
from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerBadServerOutput
from lm_agent.logs import logger
from lm_agent.parsing import lsdyna
from lm_agent.server_interfaces.license_server_interface import LicenseServerInterface
from lm_agent.utils import run_command


class LSDynaLicenseServer(LicenseServerInterface):
    """Extract license information from LS-Dyna license server."""

    def __init__(self, license_servers: typing.List[LicenseServerSchema]):
        """Initialize the license server instance with the license server host and parser."""
        self.license_servers = license_servers
        self.parser = lsdyna.parse

    def get_commands_list(self) -> typing.List[typing.List[str]]:
        """Generate a list of commands with the available license server hosts."""

        commands_to_run = []
        for license_server in self.license_servers:
            command_line = [
                f"{settings.LSDYNA_PATH}",
                "-s",
                f"{license_server.port}@{license_server.host}",
                "-R",
            ]
            commands_to_run.append(command_line)
        return commands_to_run

    async def get_output_from_server(self):
        """Override abstract method to get output from Ls-Dyna license server."""

        # get the list of commands for each license server host
        commands_to_run = self.get_commands_list()

        # run each command in the list, one at a time, until one succeds
        for cmd in commands_to_run:
            try:
                output = await run_command(cmd)
            except CommandFailedToExecute as e:
                logger.debug(f"Command {cmd} failed to execute: {e}")
                continue

            # try the next server if the previous didn't return the expected data
            if not output:
                continue
            return output

        raise RuntimeError("None of the checks for LS-Dyna succeeded!")

    async def get_report_item(self, feature_id: int, product_feature: str):
        """Override abstract method to parse LS-Dyna license server output into License Report Item."""

        server_output = await self.get_output_from_server()
        parsed_output = self.parser(server_output)

        (_, feature) = product_feature.split(".")

        current_feature_item = parsed_output.get(feature)

        # raise exception if parser didn't output license information
        if current_feature_item is None:
            raise LicenseManagerBadServerOutput("Invalid data returned from parser.")

        report_item = LicenseReportItem(
            feature_id=feature_id,
            product_feature=product_feature,
            used=current_feature_item["used"],
            total=current_feature_item["total"],
            uses=current_feature_item["uses"],
        )

        return report_item



import sys
import logging
from pathlib import Path
from typing import Optional, Annotated

from pydantic import AnyHttpUrl, Field, confloat
from pydantic_core import ValidationError
from pydantic_settings import BaseSettings, SettingsConfigDict

from lm_agent.constants import LogLevelEnum

DEFAULT_CACHE_DIR = Path.home() / Path(".cache/license-manager")
DEFAULT_LOG_DIR = Path("/var/log/license-manager-agent")
DEFAULT_BIN_PATH = Path(__file__).parent.parent / "bin"


def _get_env_file() -> Path | None:
    """
    Determine if running in test mode and return the correct path to the .env file if not.
    """
    _test_mode = "pytest" in sys.modules
    if not _test_mode:
        default_dotenv_file_location = Path("/var/snap/jobbergate-agent/common/.env")
        if default_dotenv_file_location.exists():
            return default_dotenv_file_location
        fallback_dotenv_file_location = Path("/etc/default/license-manager-agent")
        if fallback_dotenv_file_location.exists():
            return fallback_dotenv_file_location
        return Path(".env")
    return None


class Settings(BaseSettings):
    """
    App config.

    If you are setting these in the environment, you must prefix "LM_AGENT_", e.g.
    LM_AGENT_LOG_LEVEL=DEBUG
    """

    # Deployment environment
    DEPLOY_ENV: str = "LOCAL"

    # Log level
    LOG_LEVEL: LogLevelEnum = LogLevelEnum.INFO

    # Base URL of the License Manager API
    BACKEND_BASE_URL: AnyHttpUrl = Field("http://127.0.0.1:8000")

    # Location of the log directory
    LOG_BASE_DIR: Optional[Path] = DEFAULT_LOG_DIR

    # Token cache directory
    CACHE_DIR: Path = DEFAULT_CACHE_DIR

    # Path to Slurm binaries
    SCONTROL_PATH: Path = Path("/usr/bin/scontrol")
    SACCTMGR_PATH: Path = Path("/usr/bin/sacctmgr")
    SQUEUE_PATH: Path = Path("/usr/bin/squeue")

    # Path to the binary for lmutil (needed for FlexLM licenses)
    LMUTIL_PATH: Path = DEFAULT_BIN_PATH / "lmutil"

    # Path to the binary for rlmutil (needed for RLM licenses)
    RLMUTIL_PATH: Path = DEFAULT_BIN_PATH / "rlmutil"

    # Path to the binary for lstc_qrun (needed for LS-Dyna licenses)
    LSDYNA_PATH: Path = DEFAULT_BIN_PATH / "lstc_qrun"

    # Path to the binary for lmxendutil (needed for LM-X licenses)
    LMXENDUTIL_PATH: Path = DEFAULT_BIN_PATH / "lmxendutil"

    # Path to the binary for olixtool (needed for OLicense licenses)
    OLIXTOOL_PATH: Path = DEFAULT_BIN_PATH / "olixtool"

    # Path to the binary for DSLicSrv (needed for DSLS licenses)
    DSLICSRV_PATH: Path = DEFAULT_BIN_PATH / "DSLicSrv"

    # Reservation name for reconciliation
    RESERVATION_IDENTIFIER: str = "license-manager-reservation"

    # License Manager user name to create reservations
    LM_USER: str = "license-manager"

    # Sentry specific settings
    SENTRY_DSN: Optional[str] = None
    SENTRY_TRACES_SAMPLE_RATE: Annotated[float, confloat(gt=0, le=1.0)] = 0.01
    SENTRY_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.25
    SENTRY_PROFILING_SAMPLE_RATE: Annotated[float, confloat(gt=0.0, le=1.0)] = 0.01

    # OIDC config for machine-to-machine security
    OIDC_DOMAIN: str
    OIDC_CLIENT_ID: str
    OIDC_CLIENT_SECRET: str
    OIDC_USE_HTTPS: bool = True

    # If set to `True`, reconcile will be triggered by Prolog/Epilog. Set to `False` to disable this.
    USE_RECONCILE_IN_PROLOG_EPILOG: bool = True

    # Stat interval used to report the cluster status to the API
    STAT_INTERVAL: int = 60

    # Timeout for the license server binaries
    TOOL_TIMEOUT: int = 6  # seconds

    # Encoding used for decoding the output of the license server binaries
    ENCODING: str = "utf-8"

    model_config = SettingsConfigDict(env_prefix="LM_AGENT_", env_file=_get_env_file())


def init_settings() -> Settings:
    """
    Build SETTINGS, and offer a way to gracefully fail if required settings cannot be loaded
    """
    try:
        # ignoring call-arg because we expect the parameters to
        # be set either by the environment or the dotenv file
        return Settings()  # type: ignore[call-arg]
    except ValidationError as e:
        logger = logging.getLogger("lm_agent.config")
        logger.critical(f"Failed to load settings: {str(e)}")
        sys.exit(1)


settings = init_settings()



import asyncio
import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration

from lm_agent.backend_utils.utils import check_backend_health, report_cluster_status
from lm_agent.config import settings
from lm_agent.logs import init_logging, logger
from lm_agent.scheduler import scheduler
from lm_agent.services.reconciliation import reconcile


if settings.SENTRY_DSN:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        sample_rate=settings.SENTRY_SAMPLE_RATE,
        profiles_sample_rate=settings.SENTRY_PROFILING_SAMPLE_RATE,
        traces_sample_rate=settings.SENTRY_TRACES_SAMPLE_RATE,
        environment=settings.DEPLOY_ENV,
        integrations=[
            LoggingIntegration(
                level=logging.INFO,
                event_level=logging.CRITICAL,
            ),
        ],
    )


async def scheduled_tasks():
    """
    The scheduled tasks to be run by the agent.
    """
    await check_backend_health()
    await report_cluster_status()
    await reconcile()


def main():
    """
    Main function to run the scheduled tasks.
    """
    init_logging("license-manager-agent")
    logger.info("Starting License Manager Agent")

    scheduler.start()
    scheduler.add_job(scheduled_tasks)

    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        logger.info("Stopping License Manager Agent")
        scheduler.stop()


if __name__ == "__main__":
    main()



from apscheduler.schedulers.asyncio import AsyncIOScheduler

from lm_agent.config import settings


class Scheduler:
    def __init__(self, interval=settings.STAT_INTERVAL):
        self.scheduler = AsyncIOScheduler()
        self.interval = interval

    def add_job(self, func):
        self.scheduler.add_job(func, "interval", seconds=self.interval)

    def start(self):
        self.scheduler.start()

    def stop(self):
        self.scheduler.shutdown()


scheduler = Scheduler()



"""
Invoke license stat tools to build a view of license token counts.
"""
import asyncio
import typing

from lm_agent.models import ConfigurationSchema, LicenseReportItem
from lm_agent.backend_utils.utils import get_cluster_configs_from_backend, make_feature_update
from lm_agent.exceptions import LicenseManagerNonSupportedServerTypeError, LicenseManagerEmptyReportError
from lm_agent.logs import logger
from lm_agent.server_interfaces.flexlm import FlexLMLicenseServer
from lm_agent.server_interfaces.lmx import LMXLicenseServer
from lm_agent.server_interfaces.lsdyna import LSDynaLicenseServer
from lm_agent.server_interfaces.olicense import OLicenseLicenseServer
from lm_agent.server_interfaces.rlm import RLMLicenseServer
from lm_agent.server_interfaces.dsls import DSLSLicenseServer
from lm_agent.workload_managers.slurm.cmd_utils import get_all_product_features_from_cluster


def get_local_license_configurations(
    license_configurations: typing.List[ConfigurationSchema], local_licenses: typing.List[str]
) -> typing.List[ConfigurationSchema]:
    """
    Return the license configurations from the backend that are configured on the cluster.
    """
    filtered_entries = []

    for entry in license_configurations:
        for feature in entry.features:
            feature_name = feature.name
            product_name = feature.product.name
            if f"{product_name}.{feature_name}" in local_licenses:
                filtered_entries.append(entry)
                break
    return filtered_entries


async def report() -> typing.List[LicenseReportItem]:
    """
    Get stat counts using a license stat tool.

    This function iterates over the available license_servers and associated
    features configured via LICENSE_SERVER_FEATURES and generates
    a report by requesting license information from the license_server_type.

    The return from the license server is used to reconcile license-manager's
    view of what features are available with what actually exists in the
    license server database.
    """
    report_items = []
    get_report_awaitables = []
    product_features_awaited = []

    # Get cluster configuration
    license_configurations = await get_cluster_configs_from_backend()

    local_licenses = await get_all_product_features_from_cluster()
    filtered_entries = get_local_license_configurations(license_configurations, local_licenses)

    logger.debug("#### Getting reconciliation report ####")
    logger.debug("### Licenses in the backend: ")
    logger.debug(license_configurations)
    logger.debug("### Licenses in the cluster: ")
    logger.debug(filtered_entries)

    server_type_map = dict(
        flexlm=FlexLMLicenseServer,
        rlm=RLMLicenseServer,
        lsdyna=LSDynaLicenseServer,
        lmx=LMXLicenseServer,
        olicense=OLicenseLicenseServer,
        dsls=DSLSLicenseServer,
    )

    for entry in filtered_entries:
        product_features_to_check = []
        for feature in entry.features:
            feature_info = (feature.id, f"{feature.product.name}.{feature.name}")
            product_features_to_check.append(feature_info)

        logger.debug("### Features to check: ")
        logger.debug(product_features_to_check)

        server_type = server_type_map.get(entry.type)

        if server_type is None:
            raise LicenseManagerNonSupportedServerTypeError("License server type not supported.")

        license_server_interface = server_type(entry.license_servers)

        for feature_info_to_check in product_features_to_check:
            feature_id, product_feature = feature_info_to_check

            get_report_awaitables.append(
                license_server_interface.get_report_item(feature_id, product_feature)
            )
            product_features_awaited.append(feature_info_to_check)

    results: list[BaseException | LicenseReportItem] = await asyncio.gather(
        *get_report_awaitables, return_exceptions=True
    )

    for result, feature_info in zip(results, product_features_awaited):
        feature_id, product_feature = feature_info

        if isinstance(result, Exception):
            # If the report for a feature failed, the total will set to 0, preventing jobs from running
            logger.error(f"#### Report for feature {product_feature} failed with: {str(result)} ####")

            failed_report_item = LicenseReportItem(
                feature_id=feature_id,
                product_feature=product_feature,
                used=0,
                total=0,
                uses=[],
            )
            report_items.append(failed_report_item)
            continue
        assert isinstance(result, LicenseReportItem)
        report_items.append(result)

    logger.debug("#### Reconciliation items:")
    logger.debug(report_items)

    return report_items


async def update_features() -> typing.List[LicenseReportItem]:
    """Send the license data collected from the cluster to the backend."""
    license_report = await report()

    if not license_report:
        logger.critical(
            "No license data could be collected, check that tools are installed "
            "correctly and the right hosts/ports are configured in settings"
        )
        raise LicenseManagerEmptyReportError("Got an empty response from the license server")

    features_to_update = []

    for license in license_report:
        product, feature = license.product_feature.split(".")

        feature_data = {
            "product_name": product,
            "feature_name": feature,
            "total": license.total,
            "used": license.used,
        }

        features_to_update.append(feature_data)

    await make_feature_update(features_to_update)

    return license_report
